# (C) 2025 GoodData Corporation

import logging

import pytest

from gooddata_pipelines.logger.logger import LogObserver
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceFullLoad,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_validator import (
    WorkspaceDataValidator,
)
from gooddata_pipelines.provisioning.utils.exceptions import (
    WorkspaceException,
)

# Set up logging for the test
logger = logging.getLogger("test_logger")

observer = LogObserver()
observer.subscribe(logger)


@pytest.fixture
def validator(mock_gooddata_api):
    return WorkspaceDataValidator(mock_gooddata_api)


def test_check_basic_integrity_on_valid_data(validator) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="parent_id_1",
            workspace_id="workspace_id_1",
            workspace_name="workspace_name_1",
            workspace_data_filter_id="wdf_id_1",
            workspace_data_filter_values=["wdf_values_1"],
        ),
        WorkspaceFullLoad(
            parent_id="parent_id_2",
            workspace_id="workspace_id_2",
            workspace_name="workspace_name_2",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
        WorkspaceFullLoad(
            parent_id="parent_id_3",
            workspace_id="workspace_id_3",
            workspace_name="workspace_name_3",
            workspace_data_filter_id="wdf_id_1",
            workspace_data_filter_values=["wdf_values_1"],
        ),
        WorkspaceFullLoad(
            parent_id="parent_id_3",
            workspace_id="workspace_id_3",
            workspace_name="workspace_name_3",
            workspace_data_filter_id="wdf_id_2",
            workspace_data_filter_values=["wdf_values_1", "wdf_values_2"],
        ),
    ]

    parent_workspaces, parent_wdf_map = validator._check_basic_integrity(
        source_group
    )

    assert parent_workspaces == {"parent_id_1", "parent_id_2", "parent_id_3"}
    assert parent_wdf_map == {
        "parent_id_1": ["wdf_id_1"],
        "parent_id_3": ["wdf_id_1", "wdf_id_2"],
    }


def test_check_basic_integrity_missing_parent_id(validator) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="",
            workspace_id="workspace_id_1",
            workspace_name="workspace_name_1",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
    ]

    try:
        validator._check_basic_integrity(source_group)
    except Exception as e:
        assert isinstance(e, WorkspaceException)
        assert e.error_message == "Parent ID is not defined in source data."
        assert e.workspace_id == "workspace_id_1"


def test_check_basic_integrity_missing_workspace_id(validator) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="parent_id_1",
            workspace_id="",
            workspace_name="workspace_name_1",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
    ]

    try:
        validator._check_basic_integrity(source_group)
    except Exception as e:
        assert isinstance(e, WorkspaceException)
        assert (
            e.error_message
            == "Workspace ID is not defined for parent parent_id_1"
        )


def test_check_basic_integrity_missing_wdf_id(validator) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="parent_id_1",
            workspace_id="workspace_id_1",
            workspace_name="workspace_name_1",
            workspace_data_filter_id=None,
            workspace_data_filter_values=["some_values"],
        ),
    ]

    try:
        validator._check_basic_integrity(source_group)
    except Exception as e:
        assert isinstance(e, WorkspaceException)
        assert (
            e.error_message
            == "WDF values are provided but WDF ID is not defined."
        )
        assert e.workspace_name == "workspace_name_1"
        assert e.workspace_id == "workspace_id_1"
        assert e.wdf_values == "some_values"


def test_check_basic_integrity_missing_wdf_values(validator) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="parent_id_1",
            workspace_id="workspace_id_1",
            workspace_name="workspace_name_1",
            workspace_data_filter_id="wdf_id_1",
            workspace_data_filter_values=None,
        ),
    ]

    try:
        validator._check_basic_integrity(source_group)
    except Exception as e:
        assert isinstance(e, WorkspaceException)
        assert (
            e.error_message
            == "WDF ID is defined but no WDF values are provided"
        )
        assert e.workspace_id == "workspace_id_1"
        assert e.wdf_id == "wdf_id_1"
        assert e.workspace_name == "workspace_name_1"
        assert e.wdf_values is None


def test_check_basic_integrity_missing_both_ids(validator) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="",
            workspace_id="",
            workspace_name="",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
    ]

    try:
        validator._check_basic_integrity(source_group)
    except Exception as e:
        assert isinstance(e, WorkspaceException)
        assert e.error_message == (
            "Parent ID and workspace ID are not defined for at least one row. Please check the source data."
        )


def test_check_basic_integrity_duplicates_warning(validator, caplog) -> None:
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="parent_id_1",
            workspace_id="workspace_id_1",
            workspace_name="workspace_name_1",
            workspace_data_filter_id="wdf_id_1",
            workspace_data_filter_values=["wdf_values_1"],
        ),
        WorkspaceFullLoad(
            parent_id="parent_id_1",
            workspace_id="workspace_id_1",
            workspace_name="workspace_name_1",
            workspace_data_filter_id="wdf_id_1",
            workspace_data_filter_values=["wdf_values_1"],
        ),
    ]
    with caplog.at_level("WARNING"):
        validator._check_basic_integrity(source_group)

    warning_log_found = any(
        "Duplicate combinations of parent_id, workspace_id, wdf_id exist in the source data."
        in message
        for message in caplog.messages
    )

    assert warning_log_found, (
        "Warning log for duplicate combinations not found in logs."
    )



# (C) 2025 GoodData Corporation

import attrs
import requests

from gooddata_pipelines.api import GoodDataApi
from gooddata_pipelines.api.gooddata_api import API_VERSION
from gooddata_pipelines.backup_and_restore.csv_reader import CSVReader
from gooddata_pipelines.backup_and_restore.models.input_type import InputType
from gooddata_pipelines.backup_and_restore.models.workspace_response import (
    Workspace,
    WorkspaceResponse,
)
from gooddata_pipelines.logger import LogObserver


class BackupInputProcessor:
    """Class to handle the input CSV and prepare the actual input for the backup.

    Based on the InputType value, this class will determine which approach to take
    in getting the IDs of workspaces to backup. It will then call appropriate
    GoodData Cloud endpoints to get the IDs and return them as a list.
    """

    _api: GoodDataApi
    base_workspace_endpoint: str
    hierarchy_endpoint: str
    all_workspaces_endpoint: str

    def __init__(self, api: GoodDataApi, page_size: int) -> None:
        self._api = api
        self.page_size = page_size
        self.logger = LogObserver()
        self.csv_reader = CSVReader()

        self.set_endpoints()

    def set_endpoints(self) -> None:
        """Sets the hierarchy endpoint for the API client."""
        self.base_workspace_endpoint = "/api/v1/entities/workspaces"
        self.hierarchy_endpoint = (
            f"{self.base_workspace_endpoint}?"
            + "filter=parent.id=={parent_id}"
            + f"&include=parent&page=0&size={self.page_size}&sort=name,asc&metaInclude=page,hierarchy"
        )
        self.all_workspaces_endpoint = f"{self.base_workspace_endpoint}?page=0&size={self.page_size}&sort=name,asc&metaInclude=page"

    @attrs.define
    class _ProcessDataOutput:
        workspace_ids: list[str]
        sub_parents: list[str] | None = None

    def fetch_page(self, url: str) -> WorkspaceResponse:
        """Fetch a page of workspaces."""

        # Separate the API path from the URL so that it can be fed to the api class
        endpoint: str = url.split(f"api/{API_VERSION}")[1]
        response: requests.Response = self._api._get(endpoint)
        if response.ok:
            return WorkspaceResponse(**response.json())
        else:
            raise RuntimeError(
                f"Failed to fetch data from the API. URL: {endpoint}"
            )

    @staticmethod
    def process_data(data: list[Workspace]) -> _ProcessDataOutput:
        """Extract children and sub-parents from workspace data."""
        children: list[str] = []
        sub_parents: list[str] = []

        for workspace in data:
            # append child workspace IDs
            children.append(workspace.id)

            # if hierarchy is present and has children, append child workspace ID to sub_parents
            if workspace.meta and workspace.meta.hierarchy:
                if workspace.meta.hierarchy.children_count > 0:
                    sub_parents.append(workspace.id)
        return BackupInputProcessor._ProcessDataOutput(children, sub_parents)

    def log_paging_progress(self, response: WorkspaceResponse) -> None:
        """Log the progress of paging through API responses if paginatino data is present"""
        current_page: int | None
        total_pages: int | None

        if response.meta.page:
            current_page = response.meta.page.number + 1
            total_pages = response.meta.page.total_pages
        else:
            current_page = None
            total_pages = None

        if current_page and total_pages:
            self.logger.info(f"Fetched page: {current_page} of {total_pages}")

    def _paginate(
        self, url: str | None
    ) -> list["BackupInputProcessor._ProcessDataOutput"]:
        """Paginates through the API responses and returns a list of workspace IDs."""
        result: list[BackupInputProcessor._ProcessDataOutput] = []
        while url:
            response: WorkspaceResponse = self.fetch_page(url)
            self.log_paging_progress(response)
            result.append(self.process_data(response.data))
            url = response.links.next

        return result

    def get_hierarchy(self, parent_id: str) -> list[str]:
        """Returns a list of workspace IDs in the hierarchy."""
        self.logger.info(f"Fetching children of {parent_id}")
        url = self.hierarchy_endpoint.format(parent_id=parent_id)

        all_children, sub_parents = [], []

        results: list[BackupInputProcessor._ProcessDataOutput] = self._paginate(
            url
        )

        for result in results:
            all_children.extend(result.workspace_ids)
            if result.sub_parents:
                sub_parents.extend(result.sub_parents)

        for subparent in sub_parents:
            all_children += self.get_hierarchy(subparent)

        if not all_children:
            self.logger.warning(
                f"No child workspaces found for parent workspace ID: {parent_id}"
            )

        return all_children

    def get_all_workspaces(self) -> list[str]:
        """Returns a list of all workspace IDs in the organization."""
        # TODO: can be optimized - requests can be sent asynchronously.
        # Use the total number of pages to calculate the number of requests
        # to be sent. Use semaphore or otherwise limit the number of concurrent
        # requests to avoid putting too much load on the server.
        self.logger.info("Fetching all workspaces")
        url = self.all_workspaces_endpoint

        all_workspaces: list[str] = []

        results: list[BackupInputProcessor._ProcessDataOutput] = self._paginate(
            url
        )

        for result in results:
            all_workspaces.extend(result.workspace_ids)

        if not all_workspaces:
            self.logger.warning("No workspaces found in the organization.")

        return all_workspaces

    def get_ids_to_backup(
        self,
        input_type: InputType,
        path_to_csv: str | None = None,
        workspace_ids: list[str] | None = None,
    ) -> list[str]:
        """Returns the list of workspace IDs to back up based on the input type."""

        if input_type in (InputType.LIST_OF_WORKSPACES, InputType.HIERARCHY):
            if (path_to_csv is None) == (workspace_ids is None):
                raise ValueError(
                    f"Path to CSV and list of workspace IDs must be specified exclusively for this input type: {input_type.value}"
                )

            # If we're backing up based on the list, simply read it from the CSV
            list_of_parents = []
            if path_to_csv is not None:
                list_of_parents = self.csv_reader.read_backup_csv(path_to_csv)
            if workspace_ids is not None:
                list_of_parents = workspace_ids

            if input_type == InputType.LIST_OF_WORKSPACES:
                return list_of_parents
            else:
                # For hierarchy backup, we read the CSV and treat it as a list of
                # parent workspace IDs. Then we retrieve the children of each parent,
                # including their children, and so on. The parent workspaces are
                # also included in the backup.
                list_of_children: list[str] = []

                for parent in list_of_parents:
                    list_of_children.extend(self.get_hierarchy(parent))

                return list_of_parents + list_of_children

        # If we're backing up the entire organization, we simply get all workspaces
        elif input_type == InputType.ORGANIZATION:
            list_of_workspaces = self.get_all_workspaces()
            return list_of_workspaces

        else:
            # This path should be unreachable as long as the conditions above
            # exhaustively check all values of InputType Enum.
            raise ValueError(f"Unsupported input type: {input_type.value}")



# (C) 2025 GoodData Corporation

from gooddata_pipelines import WorkspaceProvisioner
from gooddata_pipelines.provisioning.provisioning import Provisioning


def test_create_groups_base_case(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    provisioner: Provisioning = workspace_provisioner
    test_source_ids = {"source_id_1", "source_id_2", "source_id_3"}
    test_panther_ids = {"source_id_2", "source_id_3", "source_id_4"}

    id_groups = provisioner._create_groups(test_source_ids, test_panther_ids)

    assert id_groups.ids_in_both_systems == {"source_id_2", "source_id_3"}
    assert id_groups.ids_to_delete == {"source_id_4"}
    assert id_groups.ids_to_create == {"source_id_1"}


def test_create_groups_empty_sets(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    provisioner: Provisioning = workspace_provisioner
    test_source_ids: set[str] = set()
    test_panther_ids: set[str] = set()

    id_groups = provisioner._create_groups(test_source_ids, test_panther_ids)

    assert id_groups.ids_in_both_systems == set()
    assert id_groups.ids_to_delete == set()
    assert id_groups.ids_to_create == set()


def test_create_groups_no_overlap(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    provisioner: Provisioning = workspace_provisioner
    test_source_ids = {"source_id_1", "source_id_2"}
    test_panther_ids = {"source_id_3", "source_id_4"}

    id_groups = provisioner._create_groups(test_source_ids, test_panther_ids)

    assert id_groups.ids_in_both_systems == set()
    assert id_groups.ids_to_delete == {"source_id_3", "source_id_4"}
    assert id_groups.ids_to_create == {"source_id_1", "source_id_2"}


def test_create_groups_full_overlap(
    workspace_provisioner: WorkspaceProvisioner,
) -> None:
    provisioner: Provisioning = workspace_provisioner
    test_source_ids = {"source_id_1", "source_id_2"}
    test_panther_ids = {"source_id_1", "source_id_2"}

    id_groups = provisioner._create_groups(test_source_ids, test_panther_ids)

    assert id_groups.ids_in_both_systems == {"source_id_1", "source_id_2"}
    assert id_groups.ids_to_delete == set()
    assert id_groups.ids_to_create == set()



# (C) 2025 GoodData Corporation

from enum import Enum
from typing import Annotated

import yaml
from pydantic import BaseModel, Field, model_validator

from gooddata_pipelines.backup_and_restore.constants import BackupSettings


class StorageType(Enum):
    """Type of storage."""

    S3 = "s3"
    LOCAL = "local"


class S3StorageConfig(BaseModel):
    """Configuration for S3 storage.

    Can be created using the following constructor methods:
    - `from_iam_role`
    - `from_aws_credentials`
    - `from_aws_profile`
    """

    backup_path: str
    bucket: str
    profile: str | None = None
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    aws_default_region: str = "us-east-1"

    @classmethod
    def from_iam_role(cls, backup_path: str, bucket: str) -> "S3StorageConfig":
        """Use default IAM role or environment credentials.

        Args:
            backup_path: The path to the backup directory.
            bucket: The name of the S3 bucket.

        Returns:
            S3StorageConfig: The S3 storage configuration.
        """
        return cls(backup_path=backup_path, bucket=bucket)

    @classmethod
    def from_aws_credentials(
        cls,
        backup_path: str,
        bucket: str,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        aws_default_region: str,
    ) -> "S3StorageConfig":
        """Use explicit AWS access keys and region.

        Args:
            backup_path: The path to the backup directory.
            bucket: The name of the S3 bucket.
            aws_access_key_id: The AWS access key ID.
            aws_secret_access_key: The AWS secret access key.
            aws_default_region: The AWS default region.

        Returns:
            S3StorageConfig: The S3 storage configuration.
        """
        return cls(
            backup_path=backup_path,
            bucket=bucket,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_default_region=aws_default_region,
        )

    @classmethod
    def from_aws_profile(
        cls, backup_path: str, bucket: str, profile: str
    ) -> "S3StorageConfig":
        """Use a named AWS CLI profile.

        Args:
            backup_path: The path to the backup directory.
            bucket: The name of the S3 bucket.
            profile: The name of the AWS profile.

        Returns:
            S3StorageConfig: The S3 storage configuration.
        """
        return cls(backup_path=backup_path, bucket=bucket, profile=profile)


class LocalStorageConfig(BaseModel):
    """Placeholder for local storage config."""

    backup_path: str = Field(default="local_backups")


class BackupRestoreConfig(BaseModel):
    """Configuration for backup and restore.

    Args:
        storage_type: The type of storage to use. Defaults to `StorageType.LOCAL`.
        storage: Storage configuration. Either `S3StorageConfig` or `LocalStorageConfig`. Defaults to `LocalStorageConfig()`.
        api_page_size: The page size for fetching workspace relationships. Defaults to `BackupSettings.API.PAGE_SIZE`.
        batch_size: The batch size for fetching workspace relationships. Defaults to `BackupSettings.API.BATCH_SIZE`.
        api_calls_per_second: The maximum API calls per second (rate limiting). Defaults to `BackupSettings.API.CALLS_PER_SECOND`.
    """

    storage_type: StorageType = Field(default=StorageType.LOCAL)
    storage: S3StorageConfig | LocalStorageConfig = Field(
        default_factory=LocalStorageConfig
    )
    api_page_size: Annotated[
        int,
        Field(
            gt=0,
            description="Page size must be greater than 0",
        ),
    ] = Field(default=BackupSettings.API.PAGE_SIZE)
    batch_size: Annotated[
        int,
        Field(
            gt=0,
            description="Batch size must be greater than 0",
        ),
    ] = Field(default=BackupSettings.API.BATCH_SIZE)
    api_calls_per_second: Annotated[
        float,
        Field(
            gt=0,
            description="Maximum API calls per second (rate limiting)",
        ),
    ] = Field(default=BackupSettings.API.CALLS_PER_SECOND)

    @classmethod
    def from_yaml(cls, conf_path: str) -> "BackupRestoreConfig":
        with open(conf_path, "r") as stream:
            conf: dict = yaml.safe_load(stream)
        return cls(**conf)

    @model_validator(mode="after")
    def validate_storage(self) -> "BackupRestoreConfig":
        """Check that the storage gets correct configuration when using S3 storage"""
        if self.storage_type == StorageType.S3:
            if not isinstance(self.storage, S3StorageConfig):
                raise ValueError(
                    "S3 storage must be configured with S3StorageConfig object"
                )
        elif self.storage_type == StorageType.LOCAL:
            if not isinstance(self.storage, LocalStorageConfig):
                raise ValueError(
                    "Local storage must be configured with LocalStorageConfig object"
                )
        return self



# (C) 2025 GoodData Corporation

"""This module defines data models for user data filters in a GoodData workspace."""

import attrs
from pydantic import BaseModel, ConfigDict


@attrs.define
class UserDataFilterGroup:
    udf_id: str
    udf_values: list[str]


@attrs.define
class WorkspaceUserDataFilters:
    workspace_id: str
    user_data_filters: list["UserDataFilterGroup"] = attrs.field(factory=list)


class UserDataFilterFullLoad(BaseModel):
    model_config = ConfigDict(extra="forbid")

    workspace_id: str
    udf_id: str
    udf_value: str


class UserDataFilterIncrementalLoad(UserDataFilterFullLoad):
    is_active: bool



# (C) 2025 GoodData Corporation

from ._version import __version__

# -------- Backup and Restore --------
from .backup_and_restore.backup_manager import BackupManager
from .backup_and_restore.models.storage import (
    BackupRestoreConfig,
    LocalStorageConfig,
    S3StorageConfig,
    StorageType,
)
from .backup_and_restore.restore_manager import (
    RestoreManager,
    WorkspaceToRestore,
)
from .backup_and_restore.storage.local_storage import LocalStorage
from .backup_and_restore.storage.s3_storage import S3Storage

# -------- LDM Extension --------
from .ldm_extension.ldm_extension_manager import LdmExtensionManager
from .ldm_extension.models.custom_data_object import (
    ColumnDataType,
    CustomDatasetDefinition,
    CustomFieldDefinition,
    CustomFieldType,
)

# -------- Provisioning --------
from .provisioning.entities.user_data_filters.models.udf_models import (
    UserDataFilterFullLoad,
)
from .provisioning.entities.user_data_filters.user_data_filters import (
    UserDataFilterProvisioner,
)
from .provisioning.entities.users.models.permissions import (
    EntityType,
    PermissionFullLoad,
    PermissionIncrementalLoad,
)
from .provisioning.entities.users.models.user_groups import (
    UserGroupFullLoad,
    UserGroupIncrementalLoad,
)
from .provisioning.entities.users.models.users import (
    UserFullLoad,
    UserIncrementalLoad,
)
from .provisioning.entities.users.permissions import PermissionProvisioner
from .provisioning.entities.users.user_groups import UserGroupProvisioner
from .provisioning.entities.users.users import UserProvisioner
from .provisioning.entities.workspaces.models import (
    WorkspaceFullLoad,
    WorkspaceIncrementalLoad,
)
from .provisioning.entities.workspaces.workspace import WorkspaceProvisioner

# -------- Generic Provisioning --------
from .provisioning.generic.config import WorkflowType
from .provisioning.generic.provision import provision

__all__ = [
    "BackupManager",
    "RestoreManager",
    "WorkspaceToRestore",
    "BackupRestoreConfig",
    "StorageType",
    "LocalStorage",
    "S3Storage",
    "WorkspaceFullLoad",
    "WorkspaceProvisioner",
    "UserIncrementalLoad",
    "UserGroupIncrementalLoad",
    "PermissionFullLoad",
    "LocalStorageConfig",
    "S3StorageConfig",
    "PermissionIncrementalLoad",
    "UserFullLoad",
    "UserGroupFullLoad",
    "UserProvisioner",
    "UserGroupProvisioner",
    "WorkspaceIncrementalLoad",
    "PermissionProvisioner",
    "UserDataFilterProvisioner",
    "UserDataFilterFullLoad",
    "EntityType",
    "LdmExtensionManager",
    "CustomDatasetDefinition",
    "CustomFieldDefinition",
    "ColumnDataType",
    "CustomFieldType",
    "provision",
    "WorkflowType",
    "__version__",
]



# (C) 2025 GoodData Corporation

"""Module for exceptions used in GoodData Pipelines provisioning."""

from gooddata_pipelines.provisioning.utils.utils import AttributesMixin


# TODO: Use the generic context exception and phase out the specific ones
#  - we don't need to conform to process-specific schema anymore
class ContextException(Exception, AttributesMixin):
    def __init__(
        self, message: str, *context_objects: object, **kwargs: str
    ) -> None:
        """Exception raised during context processing."""
        super().__init__(message)
        attributes = self.get_attrs(*context_objects, overrides=kwargs)

        for key, value in attributes.items():
            setattr(self, key, value)


class ProvisioningException(Exception, AttributesMixin):
    def __init__(
        self, message: str, *context_objects: object, **kwargs: str
    ) -> None:
        """Exception raised during provisioning."""
        super().__init__(message)
        self.attributes = self.get_attrs(*context_objects, overrides=kwargs)
        self.error_message: str = message


class WorkspaceException(ProvisioningException):
    def __init__(
        self,
        message: str,
        *context_objects: object,
        **kwargs: str,
    ) -> None:
        """Exception raised during workspace provisioning."""
        super().__init__(message, *context_objects, **kwargs)

        self.http_status: str = self.attributes.get(
            "http_status", "500 Internal Server Error"
        )
        self.http_method: str | None = self.attributes.get("http_method", "NA")
        self.workspace_id: str = self.attributes.get("workspace_id", "NA")
        self.workspace_name: str | None = self.attributes.get(
            "workspace_name", "NA"
        )
        self.wdf_id: str | None = self.attributes.get("wdf_id", None)
        self.wdf_values: str | None = self.attributes.get("wdf_values", None)
        self.api_endpoint: str = self.attributes.get(
            "api_endpoint", "workspace_provisioning"
        )


class WorkspaceDataIntegrityException(WorkspaceException):
    def __init__(
        self, message: str, *context_objects: object, **kwargs: str
    ) -> None:
        """Exception raised during workspace validation."""
        super().__init__(message, *context_objects, **kwargs)

        self.workspace_id: str = self.attributes.get("workspace_id", "NA")
        self.workspace_name: str | None = self.attributes.get(
            "workspace_name", None
        )
        self.wdf_id: str | None = self.attributes.get("wdf_id", None)
        self.wdf_values: str | None = self.attributes.get("wdf_values", None)
        self.api_endpoint: str = self.attributes.get(
            "api_endpoint", "workspace_data_validation"
        )


class BaseUserException(ProvisioningException):
    def __init__(
        self, message: str, *context_objects: object, **kwargs: str
    ) -> None:
        """Exception raised during user provisioning."""
        super().__init__(message, *context_objects, **kwargs)

        self.http_status: str = self.attributes.get(
            "http_status", "500 Internal Server Error"
        )
        self.http_method: str | None = self.attributes.get("http_method", None)
        self.workspace_id: str | None = self.attributes.get(
            "workspace_id", None
        )
        self.user_id: str | None = self.attributes.get("user_id", None)
        self.user_group_id: str | None = self.attributes.get(
            "user_group_id", None
        )
        self.api_endpoint: str = self.attributes.get(
            "api_endpoint", "user_provisioning"
        )



# (C) 2025 GoodData Corporation

from typing import Any

from gooddata_sdk.catalog.user.entity_model.user import CatalogUser
from pydantic import BaseModel, ConfigDict, Field


class UserProfile(BaseModel):
    """Minimal model of api/v1/profile response.

    Does not contain all fields from the response.
    """

    user_id: str = Field(alias="userId")


class BaseUser(BaseModel):
    """Base class containing shared user fields and functionality."""

    model_config = ConfigDict(extra="forbid")

    user_id: str
    firstname: str | None
    lastname: str | None
    email: str | None
    auth_id: str | None
    user_groups: list[str]

    @classmethod
    def _create_from_sdk_data(cls, obj: CatalogUser) -> dict[str, Any]:
        """Helper method to extract common data from SDK object."""
        if obj.attributes:
            firstname = obj.attributes.firstname
            lastname = obj.attributes.lastname
            email = obj.attributes.email
            auth_id = obj.attributes.authentication_id
        else:
            firstname = None
            lastname = None
            email = None
            auth_id = None

        return {
            "user_id": obj.id,
            "firstname": firstname,
            "lastname": lastname,
            "email": email,
            "auth_id": auth_id,
            "user_groups": [ug.id for ug in obj.user_groups],
        }

    def to_sdk_obj(self) -> CatalogUser:
        """Converts to CatalogUser SDK object."""
        return CatalogUser.init(
            user_id=self.user_id,
            firstname=self.firstname,
            lastname=self.lastname,
            email=self.email,
            authentication_id=self.auth_id,
            user_group_ids=self.user_groups,
        )


class UserFullLoad(BaseUser):
    """Input validator for full load of user provisioning."""

    @classmethod
    def from_sdk_obj(cls, obj: CatalogUser) -> "UserFullLoad":
        """Creates GDUserTarget from CatalogUser SDK object."""
        base_data = cls._create_from_sdk_data(obj)
        return cls(**base_data)


class UserIncrementalLoad(BaseUser):
    """Input validator for incremental load of user provisioning."""

    is_active: bool

    @classmethod
    def from_sdk_obj(cls, obj: CatalogUser) -> "UserIncrementalLoad":
        """Creates GDUserTarget from CatalogUser SDK object."""
        base_data = cls._create_from_sdk_data(obj)
        base_data["is_active"] = True
        return cls(**base_data)



# (C) 2025 GoodData Corporation

"""Module for utilities used in GoodData Pipelines provisioning."""

from typing import Any, cast

import attrs
from requests import Response


class AttributesMixin:
    """
    Mixin class to provide a method for getting attributes of an object which may or may not exist.
    """

    def get_attrs(
        self, *objects: object, overrides: dict[str, str] | None = None
    ) -> dict[str, str]:
        """
        Returns a dictionary of attributes from the given objects.

        Args:
            objects: The objects to get the attributes from. Special handling is implemented for
                    requests.Response, __dict__ attribute is used for general objects.
            overrides: A dictionary of attributes to override the object's attributes.
        Returns:
            dict: Returns a dictionary of the objects' attributes.
        """
        attributes: dict[str, str] = {}
        for index, context_object in enumerate(objects):
            if isinstance(context_object, str):
                attributes[f"string_context_{index}"] = context_object

            if isinstance(context_object, Response):
                # for request.Response objects, keys need to be renamed to match the log schema
                attributes.update(
                    {
                        "http_status": str(context_object.status_code),
                        "http_method": getattr(
                            context_object.request, "method", "NA"
                        ),
                        "api_endpoint": getattr(
                            context_object.request, "url", "NA"
                        ),
                    }
                )
            elif attrs.has(type(context_object)):
                for key, value in attrs.asdict(
                    cast(attrs.AttrsInstance, context_object)
                ).items():
                    self._add_to_dict(attributes, key, value)
            elif hasattr(context_object, "__dict__"):
                # Generic handling for other objects
                for key, value in context_object.__dict__.items():
                    self._add_to_dict(attributes, key, value)
            else:
                attributes[f"object_{index}"] = str(context_object)

        if overrides:
            attributes.update(overrides)

        return attributes

    def _add_to_dict(
        self, attributes: dict[str, str], key: str, value: Any
    ) -> None:
        if value is None:
            return

        if isinstance(value, list):
            attributes[key] = ", ".join(str(list_item) for list_item in value)
        else:
            attributes[key] = str(value)


@attrs.define
class EntityGroupIds:
    ids_in_both_systems: set[str]
    ids_to_delete: set[str]
    ids_to_create: set[str]



# (C) 2025 GoodData Corporation
"""Module orchestrating the custom fields logic."""

from pathlib import Path

from gooddata_sdk.sdk import GoodDataSdk
from gooddata_sdk.utils import PROFILES_FILE_PATH, profile_content

from gooddata_pipelines.api import GoodDataApi
from gooddata_pipelines.ldm_extension.input_processor import (
    LdmExtensionDataProcessor,
)
from gooddata_pipelines.ldm_extension.input_validator import (
    LdmExtensionDataValidator,
)
from gooddata_pipelines.ldm_extension.models.aliases import (
    DatasetId,
    WorkspaceId,
)
from gooddata_pipelines.ldm_extension.models.analytical_object import (
    AnalyticalObject,
    AnalyticalObjects,
)
from gooddata_pipelines.ldm_extension.models.custom_data_object import (
    CustomDataset,
    CustomDatasetDefinition,
    CustomFieldDefinition,
)
from gooddata_pipelines.logger.logger import LogObserver


class LdmExtensionManager:
    """Manager for creating custom datasets and fields in GoodData workspaces."""

    INDENT = " " * 2

    @classmethod
    def create(cls, host: str, token: str) -> "LdmExtensionManager":
        return cls(host=host, token=token)

    @classmethod
    def create_from_profile(
        cls,
        profile: str = "default",
        profiles_path: Path = PROFILES_FILE_PATH,
    ) -> "LdmExtensionManager":
        """Creates a provisioner instance using a GoodData profile file."""
        content = profile_content(profile, profiles_path)
        return cls(host=content["host"], token=content["token"])

    def __init__(self, host: str, token: str):
        self._validator = LdmExtensionDataValidator()
        self._processor = LdmExtensionDataProcessor()
        self._sdk = GoodDataSdk.create(host_=host, token_=token)
        self._api = GoodDataApi(host=host, token=token)
        self.logger = LogObserver()

    def _get_objects_with_invalid_relations(
        self, workspace_id: str
    ) -> list[AnalyticalObject]:
        """Check for invalid references in the provided analytical objects.

        This method checks if the references in the provided analytical objects
        are valid. It returns a set of analytical objects that have invalid references.

        Args:
            workspace_id (str): The ID of the workspace to check.

        Returns:
            list[AnalyticalObject]: Set of analytical objects with invalid references.
        """

        analytical_objects: list[AnalyticalObject] = (
            self._get_analytical_objects(workspace_id=workspace_id)
        )

        objects_with_invalid_relations = [
            obj
            for obj in analytical_objects
            if not obj.attributes.are_relations_valid
        ]
        return objects_with_invalid_relations

    def _get_analytical_objects(
        self, workspace_id: str
    ) -> list[AnalyticalObject]:
        """Get analytical objects in the workspace.

        This method retrieves all analytical objects (metrics, visualizations, dashboards)
        in the specified workspace and returns them as a list.

        Args:
            workspace_id (str): The ID of the workspace to retrieve objects from.

        Returns:
            list[AnalyticalObject]: List of analytical objects in the workspace.
        """
        metrics_response = self._api.get_all_metrics(workspace_id)
        visualizations_response = self._api.get_all_visualization_objects(
            workspace_id
        )
        dashboards_response = self._api.get_all_dashboards(workspace_id)

        self._api.raise_if_response_not_ok(
            metrics_response,
            visualizations_response,
            dashboards_response,
        )
        metrics = AnalyticalObjects(**metrics_response.json())
        visualizations = AnalyticalObjects(**visualizations_response.json())
        dashboards = AnalyticalObjects(**dashboards_response.json())

        return metrics.data + visualizations.data + dashboards.data

    @staticmethod
    def _new_ldm_does_not_invalidate_relations(
        current_invalid_relations: list[AnalyticalObject],
        new_invalid_relations: list[AnalyticalObject],
    ) -> bool:
        """Check if the new LDM does not invalidate any new relations.

        This method compares the lists of analytical objects containing invalid
        relations. It creates sets of object IDs for each list and compares them.

        If the set of new invalid relations is a subset of the set of current
        invalid relations (that is before the changes to the LDM), the new LDM
        does not invalidate any new relations and `True` is returned.

        If the set of new invalid relations is not a subset of the current one,
        it means that the new LDM invalidates new relations and `False` is returned.

        Args:
            current_invalid_relations (list[AnalyticalObject]): The current (before
                changes to LDM) invalid relations.
            new_invalid_relations (list[AnalyticalObject]): The new (after changes to
                LDM) invalid relations.

        Returns:
            bool: True if the new LDM does not invalidate any relations, False otherwise.
        """
        # Create a set of IDs for each group, then compare those sets
        set_current_invalid_relations = {
            obj.id for obj in current_invalid_relations
        }
        set_new_invalid_relations = {obj.id for obj in new_invalid_relations}

        # If the set of new invalid relations is a subset of the current one,
        return set_new_invalid_relations.issubset(set_current_invalid_relations)

    def _process_with_relations_check(
        self,
        validated_data: dict[WorkspaceId, dict[DatasetId, CustomDataset]],
    ) -> None:
        """Check whether relations of analytical objects are valid before and after
        updating the LDM in the GoodData workspace.
        """
        # Iterate through the workspaces.
        for workspace_id, datasets in validated_data.items():
            self.logger.info(f"⚙️ Processing workspace {workspace_id}...")
            # Get current workspace layout
            current_layout = (
                self._sdk.catalog_workspace.get_declarative_workspace(
                    workspace_id
                )
            )
            # Get a set of objects with invalid relations from current workspace state
            current_invalid_relations = (
                self._get_objects_with_invalid_relations(
                    workspace_id=workspace_id
                )
            )

            # Put the LDM with custom datasets into the GoodData workspace.
            self._sdk.catalog_workspace_content.put_declarative_ldm(
                workspace_id=workspace_id,
                ldm=self._processor.datasets_to_ldm(datasets),
            )

            # Get a set of objects with invalid relations from the new workspace state
            new_invalid_relations = self._get_objects_with_invalid_relations(
                workspace_id=workspace_id
            )

            if self._new_ldm_does_not_invalidate_relations(
                current_invalid_relations, new_invalid_relations
            ):
                self._log_success_message(workspace_id)
                continue

            self.logger.error(
                f"❌ Difference in invalid relations found in workspace {workspace_id}."
            )
            self._log_diff_invalid_relations(
                current_invalid_relations, new_invalid_relations
            )

            self.logger.info(
                f"{self.INDENT}⚠️ Reverting the workspace layout to the original state."
            )
            # Put the original workspace layout back to the workspace
            try:
                self._sdk.catalog_workspace.put_declarative_workspace(
                    workspace_id=workspace_id, workspace=current_layout
                )
            except Exception as e:
                self.logger.error(
                    f"Failed to revert workspace layout in {workspace_id}: {e}"
                )

    def _log_diff_invalid_relations(
        self,
        current_invalid_relations: list[AnalyticalObject],
        new_invalid_relations: list[AnalyticalObject],
    ) -> None:
        """Logs objects with newly invalid relations.

        Objects which previously did not have invalid relations, but do so after
        updating the LDM, are logged.
        """
        # TODO: test !
        diff_to_log: list[str] = []
        for obj in new_invalid_relations:
            if obj not in current_invalid_relations:
                diff_to_log.append(
                    f"{self.INDENT}∙ {obj.id} ({obj.type}) {obj.attributes.title}"
                )
        joined_diff_to_log = "\n".join(diff_to_log)
        error_message = f"{self.INDENT}Objects with newly invalidated relations:\n{joined_diff_to_log}"

        self.logger.error(error_message)

    def _process_without_relations_check(
        self,
        validated_data: dict[WorkspaceId, dict[DatasetId, CustomDataset]],
    ) -> None:
        """Update the LDM in the GoodData workspace without checking relations."""
        for workspace_id, datasets in validated_data.items():
            # Put the LDM with custom datasets into the GoodData workspace.
            self._sdk.catalog_workspace_content.put_declarative_ldm(
                workspace_id=workspace_id,
                ldm=self._processor.datasets_to_ldm(datasets),
            )
            self._log_success_message(workspace_id)

    def _log_success_message(self, workspace_id: str) -> None:
        """Log a success message after updating the workspace LDM."""
        self.logger.info(f"✅ LDM in {workspace_id} updated successfully.")

    def process(
        self,
        custom_datasets: list[CustomDatasetDefinition],
        custom_fields: list[CustomFieldDefinition],
        check_relations: bool = True,
    ) -> None:
        """Create custom datasets and fields in GoodData workspaces.

        Creates custom datasets and fields to extend the Logical Data Model (LDM)
        in GoodData workspaces based on the provided raw data definitions. The raw
        data is validated by Pydantic models (CustomDatasetDefinition and CustomFieldDefinition).
        The defined datasets and fields are then uploaded to GoodData Cloud.

        Args:
            custom_datasets (list[CustomDatasetDefinition]): List of custom dataset definitions.
            custom_fields (list[CustomFieldDefinition]): List of custom field definitions.
            check_relations (bool): If True, checks for invalid relations in the workspace
                after updating the LDM. If the number of invalid relations increases,
                the LDM will be reverted to its previous state. If False, the check
                is skiped and the LDM is updated directly. Defaults to True.

        Raises:
            ValueError: If there are validation errors in the dataset or field definitions.
        """
        # Validate raw data and aggregate the custom field and dataset
        # definitions per workspace.
        validated_data: dict[WorkspaceId, dict[DatasetId, CustomDataset]] = (
            self._validator.validate(custom_datasets, custom_fields)
        )

        if check_relations:
            # Process the validated data with relations check.
            self._process_with_relations_check(validated_data)
        else:
            self._process_without_relations_check(validated_data)
